//
// Patron pour la generation d'une classe derivee dans ROOT
// FFasterDataReader derive de FBlockDataReader
//

#ifndef FFasterDataReader_h
#define FFasterDataReader_h

#include "FBlockDataReader.h"
#include "FParamBrut.h"

// Pour la gestion des trames Faster
const Int_t headerDataLoadSize = 12; // 12 octets d'entete de trame data
//const Int_t maxLoadOscTrame = 704;  // Nombre max de data de type oscillo dans une trame:
const Int_t maxLoadOscTrame = 700;  // NEW GEDEK !! Nombre max de data de type oscillo dans une trame:
// Normalement ETHERNET_MTU-UDP vaut 1500, mais la limite est fixee ici par le GEDEK a 1456 !!
// (GEDEK-UDP_header-Faster_Header_Trame-headerDataLoadSize-samplingMetaDataHeader)/2 = (1456-8-12-28)/2 = 704
const Int_t samplingMetaDataHeader = 28; // 28 octets en entete des donnees Sampling 
// En entete: 7 mots x 32bits = 7 mots x 4 x 8bits = 28 octets

// Gestion du nombre de points "sampler": de 2 à 718
const Int_t maxLoadSamplerTrame = 718;  // Nombre max de data de type Sampler dans une trame

// Champ de bits pour donnee de type QDC
typedef struct qdc_data{
  signed   value     : 31; // qdc value
  unsigned saturated :  1; // saturated bit
} qdc_data;


// Champ de bits pour donnee de type ADC
// 1er mot avec delatT au maximum
typedef struct adc1_data{
  unsigned pad1	     :  10; // [0-9] not used
  signed   delta_t   :  16; // saturated bit
  unsigned pad2      :   6; // [31-26] not used
} adc1_data;
// 2eme mot avec valeur ADC, bit de saturation et bit de pileup
typedef struct adc2_data{
  signed   value     : 22; // adc value
  unsigned pad	     :  8; // [22-29] not used
  unsigned saturated :  1; // saturated bit
  unsigned pileup    :  1; // pileup bit
} adc2_data;

// Champ de bits pour donnee de type ADC_HR (SP3)
typedef struct adc_hr_data{
  signed   value     : 23; // [0-22] max value
  unsigned precision :  6; // [28-23] precision
  unsigned pileup    :  1; // [29] pileup bit
  unsigned saturated :  1; // [30] saturated bit
  unsigned pad       :  1; // [31] not used (SatCPZ for test)
} adc_hr_data;

// Champ de bits pour donnee de type Periode du module RF
typedef struct period_data{
  unsigned value     : 31; // period value
  unsigned saturated :  1; // saturated bit
} period_data;

// Champ de bits pour donnee de type Electrometre
typedef struct electrometer_data{
  unsigned value     : 31; // electrometer value
  unsigned saturated :  1; // saturated bit
} electrometer_data;

// Champs de bits pour donnee de type Scaler
// 1er mot avec maximum du pulse, nombre de quanta et bit de saturation
typedef struct scaler1_data{
  signed   pulse_max : 20; // max pulse value
  unsigned pad	     :  7; // [20-26] not used
  unsigned nb_quanta :  4; // number of quanta
  unsigned saturated :  1; // saturated bit
} scaler1_data;
// 2eme mot avec largeur du pulse et position du maximum seuil à seuil
typedef struct scaler2_data{
  unsigned pulse_width : 16; // pulse width
  unsigned max_pos     : 16; // maximum position threshold to threshold
} scaler2_data;

// Champs de bits pour donnee de type QTDC (creative)
// 1er mot avec largeur seuil à seuil, precision en temps,
// bits d'information et nombre de charges 
typedef struct qtdc1_data{
  unsigned width_th2th    : 16; // width thresold to threshold
  signed   time_precision :  9; // time precision with sign bit [24]
  unsigned pileup         :  1; // pileup bit
  unsigned saturated      :  1; // saturated bit
  unsigned nbQ            :  3; // number of QDC charges
  unsigned qth2th_info    :  1; // QDC threshold to threshold information bit
  unsigned maxth2th_info  :  1; // maximum threshold to threshold information bit
} qtdc1_data;
// Dernier mot (4ème) avec position et valeur du maximum seuil à seuil
typedef struct qtdc2_data{
  unsigned max_pos : 16; // maximum position threshold to threshold
  signed   max_val : 14; // maximum value threshold to threshold
  unsigned pad     :  2; // [30-31] not used
} qtdc2_data;

// Champ de bits pour donnee de type Haute Tension
typedef struct hv_data{
  unsigned state :  4; // [3-0] State_t
  unsigned i_mon : 12; // [15-4] intensity value
  unsigned v_mon : 12; // [27-16] tension value
  unsigned pad1  :  4; // [31-28] not used
  unsigned polar :  1; // [32] polarity
  unsigned pad2	 :  7; // [39-33] not used
  unsigned vmax  :  8; // [47-40] max tension
  unsigned pad3  :  4; // [51-48] not used
  unsigned temp  : 12; // [63-52] temperature value
} hv_data;

// Champs de bits pour donnee de type QT2T
// 1er mot avec la charge
typedef struct qt2t1_data{
  signed value       : 31; // charge value (threshold to threshold)
  unsigned saturated :  1; // saturated bit
} qt2t1_data;
// 2eme mot avec la charge avant threshold et l'amplitude max
typedef struct qt2t2_data{
  signed q_baseline : 18; // charge before threshold
  unsigned max_amp  : 14; // amplitude max (after trigger)
} qt2t2_data;
// 3eme mot avec largeur et position seuil à seuil
typedef struct qt2t3_data{
  unsigned width_th2th : 16; // width thresold to threshold
  unsigned max_pos     : 16; // maximum position threshold to threshold
} qt2t3_data;

// Definition des facteurs de conversion en temps (tous exprimes en ns par defaut dans FASTER)
#define FASTER_TS2NS 2.0   	        // Timestamp to nanoseconds
#define ADC_DT2NS    8.0	        // ADC Delta_t to nanoseconds
#define ADC_P2NS     31.25e-3	        // ADC_HR Precision to nanoseconds [2ns/64) -- 2 puissance 6 bits = 64
#define QDC_TDC2NS   7.8125e-3	        // QDC TDC to nanoseconds [2ns/256]
#define RF_P2NS	     1.9073486328125e-6	// RF Period to nanoseconds [2ns/pow(2,19)]
#define RF_DT2NS     7.8125e-3		// RF Delta_t for Trig and PLL to nanoseconds [2ns/256]
#define SCALER_PW2NS 2.0		// Scaler Pulse Width to nanoseconds
#define QTDC_DT2NS   7.8125e-3		// QTDC time precision to nanoseconds [2ns/256]

// Definition des types de donnees FASTER interessantes pour RHB:
//
// FASTER System [0-9]
//--------------------
// General information types
const UShort_t SYNC_INFO    = 0;
const UShort_t START_INFO   = 1;
const UShort_t STOP_INFO    = 2;
const UShort_t MISSING_INFO = 3;
const UShort_t MISSED_INFO  = 4;
// Trigger/Merger error types
const UShort_t ERROR7 = 7;
const UShort_t ERROR8 = 8;
const UShort_t ERROR9 = 9;
//----------------------
//const Int_t SYST = 1;
//----------------------
// Group [10]
const UShort_t GROUP = 10;
//----------------------
// Time reference types [11-18]
const UShort_t TREF = 11;     // Special type without load (Timestamp in Data Header only)
const UShort_t TREF_TDC = 12; // Special type with a Time Precision measure in pico seconds in load (and Timestamp in Data Header too)
//----------------------
// RF module [19]
const UShort_t RF = 19;
// RF counters [20]
const UShort_t RF_COUNT = 20;
//----------------------
// Oscillograms [21]
const UShort_t OSC = 21;
//----------------------
// Sampler module [22]
const UShort_t SAMPLER = 22;
// Sampler counters [23]
const UShort_t SAMPLER_COUNT = 23;
//----------------------
// QDC System types [41-44]
//----------------------
const UShort_t QDCS = 40; // Internal family number to compute the label number for all QDC data in RHB
//----------------------
const UShort_t QDC1 = 41;
const UShort_t QDC2 = 42;
const UShort_t QDC3 = 43;
const UShort_t QDC4 = 44;
//----------------------
// QTDC (creative) [45]
const UShort_t QTDC = 45;
//----------------------
// QT2T [46]
const UShort_t QT2T = 46;
//----------------------
// QDC counters [50-60]
const UShort_t QDC_COUNT = 50;
// QTDC (creative) counters [51]
const UShort_t QTDC_COUNT = 51;
// QT2T counters [52]
const UShort_t QT2T_COUNT = 52;
//----------------------
// ADC System [61]
const UShort_t ADC1 = 61;          // OLD type used for standard ADC with CRRC4 filter : keep compatibility
const UShort_t CRRC4_SPECTRO = 61; // NEW type used !
//----------------------
// ADC User [62-69]
//  ... add them here ...
//----------------------
// ADC High Resolution (SP3) [62]
//const UShort_t ADC_TRAP = 62;     // OLD type used for advanced ADC with trapezoidal filter
const UShort_t TRAPEZ_SPECTRO = 62; // New type used !
// ADC counters [70-80]
const UShort_t ADC_COUNT = 70;
//----------------------
// Electrometer type [81]
const UShort_t ELECTRO = 81;
// Scaler type [82]
const UShort_t SCALER = 82;
// Scaler counter [83]
const UShort_t SCALER_COUNT = 83;
// High Voltage [84]
const UShort_t HIGH_VOLTAGE = 84;
//----------------------
// QDC_TDC types [141-144]
//----------------------
const UShort_t QDC_TDCS = 140; // Internal family number to compute the label number for all QDC_TDC data in RHB
//----------------------
const UShort_t QDC_TDC1 = 141;
const UShort_t QDC_TDC2 = 142;
const UShort_t QDC_TDC3 = 143;
const UShort_t QDC_TDC4 = 144;
//----------------------
// QDC_TOF types [241-244]
//----------------------
const UShort_t QDC_TOFS = 240; // Internal family number to compute the label number for all QDC_TOF data in RHB
//----------------------
const UShort_t QDC_TOF1 = 241;
const UShort_t QDC_TOF2 = 242;
const UShort_t QDC_TOF3 = 243;
const UShort_t QDC_TOF4 = 244;
// Fin definition des types de donnees FASTER


class FFasterDataReader : public FBlockDataReader {

  protected:
    FDevice *FasterDevice;
    UShort_t type;
    UShort_t label;
    ULong64_t clock;
    UShort_t load;
    TString fileParamDesc;
    Int_t cptParam;
    Int_t nerrors;
    THashList labelToName; // liste intermediaire pour accelerer l'acces aux parametres
    UShort_t magic_data;   // Magic word for FASTER data (0xAA) 

  private:
    void GenerateErrorParameters(TString namePar);
    void GenerateTRefParameters(TString namePar);
    void GenerateOscilloParameters(TString namePar);
    void GenerateAdcParameters(TString namePar);
    void GenerateAdcHrParameters(TString namePar); 
    void GenerateAdcCountParameters(TString namePar);
    void GenerateQdcCountParameters(TString namePar);
    void GenerateQdcTdcParameters(UShort_t qdcType, TString namePar, Int_t multiHit=0);
    void GenerateQdcTofParameters(UShort_t qdcTofType, TString namePar);
    void GenerateRFParameters(TString namePar);
    void GenerateRFCounterParameters(TString namePar);
    void GenerateElectrometerParameters(TString namePar);
    void GenerateScalerParameters(TString namePar);
    void GenerateScalerCounterParameters(TString namePar);
    void GenerateSamplerParameters(TString namePar);
    void GenerateSamplerCounterParameters(TString namePar);
    void GenerateQtdcParameters(TString namePar);
    void GenerateQtdcCounterParameters(TString namePar);
    void GenerateQt2tParameters(TString namePar);
    void GenerateQt2tCounterParameters(TString namePar);
    void GenerateHVParameters(TString namePar);
    void ReadError(void);
    void ReadOscilloData(void);
    void ReadOscilloMetaData(void);
    void ReadQDC(UShort_t nbQ);
    void ReadQDCCount(void);
    void ReadADC(void);
    void ReadADCHR(void);
    void ReadADCCount(void);
    void ReadQDC_TDC(UShort_t nbQ);
    void ReadQDC_TOF(UShort_t nbQ);
    void ReadRF(void);
    void ReadRFCounters(void);
    void ReadElectrometer(void);
    void ReadScaler(void);
    void ReadScalerCount(void);
    void ReadSampler(void);
    void ReadSamplerCount(void);
    void ReadQtdc(void);
    void ReadQtdcCount(void);
    void ReadQt2t(void);
    void ReadQt2tCount(void);
    void ReadHV(void);
    Double_t ComputeHvIntensity(hv_data hv);
    Double_t ComputeHvVoltage(hv_data hv);
    void ReadTimeRef(void);
    void ReadGroup(void);

  protected:
    virtual Int_t GetNextEvent(void)=0;
    void GetHeaderDataBytes(void);
    void GetData(void);

  public:
    FFasterDataReader(void);    // constructeur par defaut
    FFasterDataReader(const FFasterDataReader &ap); // constructeur par copie

    ~FFasterDataReader(void);		// desctructeur;

#if ROOT_VERSION_CODE >= ROOT_VERSION(3,4,0)
    virtual void Copy(TObject & obj) const;
#else
    virtual void Copy(TObject & obj);
#endif
    FFasterDataReader& operator = (const FFasterDataReader &hp); // operateur =

    virtual void InitFromEnv(TEnv *env);
    virtual void SetDevice(FDevice *dev=0);
    virtual void InitListOfParamsBruts(void);
    virtual void SetParamBrutsValues(void);
    virtual FParamBrut *GetParamBrut(Int_t labelRHB);
    virtual void SetParamDescFile(const Char_t *ipdf) { fileParamDesc = ipdf;}
    virtual const Char_t *GetParamDescFile(void) { return fileParamDesc.Data();}
    virtual const Char_t *GetDataReaderInfo(void) { return (const Char_t *)"";}
    virtual void ResetDataReaderInfo(void) {;}
    
    virtual void AddParam(FParamBrut *p=0); // Ajout d'un parametre aux differentes listes
    virtual void SetValue(FParamBrut *fpb, Double_t value);
    virtual FParamBrut *GenerateParamBrut(Int_t multiHit);
    
    ClassDef(FFasterDataReader,1)
};

#endif

